from __future__ import annotations

import sys
from dataclasses import dataclass
from os import environ
from pathlib import Path
from typing import Mapping, Sequence

from ._spawnutil import c


@dataclass(frozen=True)
class SparkleFrameworkInfo:
    frameworkPath: Path
    downloadURL: str
    archivePath: Path

    @classmethod
    def fromVersion(cls, version: str) -> SparkleFrameworkInfo:
        home = environ["HOME"]
        qualname = f"Sparkle-{version}"
        archive = f"{qualname}.tar.xz"
        versionDir = Path(f"{home}/.local/firstparty/sparkle-project.org/{qualname}")
        return cls(
            frameworkPath=versionDir / "Sparkle.framework",
            downloadURL=f"https://github.com/sparkle-project/Sparkle/releases/download/{version}/{archive}",
            archivePath=versionDir / archive,
        )

    async def download(self) -> None:
        """
        Download the given version of the Sparkle framework and unpack it into
        a known location (asynchronously, by spawning subprocesses using a
        Twisted reactor).
        """
        archiveDownloadPath = str(self.archivePath.parent)
        await c.mkdir("-p", archiveDownloadPath)
        await c.curl("-LO", self.downloadURL, workingDirectory=archiveDownloadPath)
        await c.tar("xf", self.archivePath.name, workingDirectory=archiveDownloadPath)


@dataclass(frozen=True)
class AppCastDeployment:
    """
    Configuration values for deploying to an appcast hosted on a web server
    that we can rsync a local directory to.

    @note: hopefully obviously, rsync-to-a-directory is not the most robust
        deployment mechanism and we should probably have other ways of managing
        the input to the appcast.
    """

    privateKeyAccount: str
    localUpdatesFolder: Path
    remoteHost: str
    remotePath: str

    async def deploy(self, fwinfo: SparkleFrameworkInfo) -> None:
        self.localUpdatesFolder.mkdir(parents=True, exist_ok=True)
        for release in Path("dist").glob("*.release.app.zip"):
            target = self.localUpdatesFolder / release.name
            if target.exists():
                raise RuntimeError(f"version conflict for {release}")
            release.rename(target)
        await c[str(fwinfo.frameworkPath.parent / "bin" / "generate_appcast")](
            "--account", self.privateKeyAccount, str(self.localUpdatesFolder)
        )
        await c.rsync(
            "-avz",
            # NB: homebrew version required, since this is a new feature as of
            # 2020 and apple always insist on core utilities being decades out
            # of date
            "--mkpath",
            "--delete",
            str(self.localUpdatesFolder).rstrip("/") + "/",
            f"{self.remoteHost}:{self.remotePath.rstrip('/')}/",
        )


@dataclass
class SparkleData:
    publicEDKey: str
    feedURL: str
    sparkleFramework: SparkleFrameworkInfo
    deployment: AppCastDeployment

    def plist(self) -> Mapping[str, str]:
        return {"SUPublicEDKey": self.publicEDKey, "SUFeedURL": self.feedURL}

    @classmethod
    def withConfig(
        cls,
        *,
        sparkleVersion: str,
        publicEDKey: str,
        feedURL: str,
        keychainAccount: str,
        localUpdatesFolder: Path,
        remoteHost: str,
        remotePath: str,
    ) -> SparkleData:
        self = cls(
            publicEDKey=publicEDKey,
            feedURL=feedURL,
            sparkleFramework=SparkleFrameworkInfo.fromVersion(sparkleVersion),
            deployment=AppCastDeployment(
                privateKeyAccount=keychainAccount,
                localUpdatesFolder=localUpdatesFolder,
                remoteHost=remoteHost,
                remotePath=remotePath,
            ),
        )
        return self

    async def deploy(self) -> None:
        await self.deployment.deploy(self.sparkleFramework)


def _prefix(p: Path, fx: str) -> Path:
    return p.parent / (fx + p.name)


@dataclass(kw_only=True)
class AppDescription:
    """
    An L{AppDescription} is a high-level description of an application.
    """

    bundleID: str
    bundleName: str
    icnsFile: Path
    mainPythonScript: Path

    dataFiles: Sequence[Path] = ()
    otherFrameworks: Sequence[Path] = ()
    dockIconAtStart: bool = True
    sparkleData: SparkleData | None = None

    def varyBundleForTesting(self) -> AppDescription:
        """
        Add 'Ci' and 'Test' variants based on the environment variables:

            1. C{CI_MODE}, which can be set for running in e.g. Github Actions
               continuous integration, and

            2. C{TEST_MODE}, for generating an interactive version of the
               bundle intended to test specific functionality.

        If one of these variables is set, a few changes will be made to the
        build configuration:

            1. The bundle's name will have either 'Ci' or 'Test' prepended to
               it.

            2. the bundle's identifier will have 'test' or 'ci' prepended to
               its last dotted segment.

            3. the bundle's script will be changed to have 'Ci' or 'Test'
               prepended (so a sibling script should exist with that name).

            4. the bundle's icon file name will have 'Test' or 'Ci' prepended
               to it, similar to the script.

            5. the bundle will not include metadata about the Sparkle
               framework, as it should not be updated or deployed publicly.
        """

        def check_mode() -> str:
            match environ:
                case {"CI_MODE": b}:
                    if b:
                        return "Ci"
                case {"TEST_MODE": b}:
                    if b:
                        return "Test"
            return ""

        mode = check_mode()
        if mode == "":
            return self

        return AppDescription(
            bundleID=".".join(
                [
                    *(segments := self.bundleID.split("."))[:-1],
                    mode + segments[-1],
                ]
            ),
            bundleName=mode + self.bundleName,
            icnsFile=_prefix(self.icnsFile, mode),
            mainPythonScript=_prefix(self.mainPythonScript, mode),
            dataFiles=self.dataFiles,
            otherFrameworks=self.otherFrameworks,
            dockIconAtStart=self.dockIconAtStart,
            sparkleData=None,
        )

    def setupOptions(self) -> dict[str, object]:
        """
        Create a collection of arguments to L{setuptools.setup}.
        """
        # Import py2app for its side-effect of registering the setuptools
        # command.
        assert __import__("py2app") is not None
        sparklePlist: Mapping[str, str] = (
            {} if self.sparkleData is None else self.sparkleData.plist()
        )
        sparkleFrameworks = (
            []
            if self.sparkleData is None
            else [str(self.sparkleData.sparkleFramework.frameworkPath)]
        )
        # resolving a virtualenv gives the actualenv
        pyVersionDir = Path(sys.executable).resolve().parent.parent

        # Tcl/Tk frameworks distributed with Python3.13+ need to be excluded.
        frameworksDir = pyVersionDir / "Frameworks"
        dylibExcludes: list[Path] = []
        if frameworksDir.is_dir():
            dylibExcludes.extend(frameworksDir.iterdir())
        infoPList = {
            "LSUIElement": not self.dockIconAtStart,
            "CFBundleIdentifier": self.bundleID,
            "CFBundleName": self.bundleName,
            # py2app probably doesn't require this any more most of the
            # time, but it doesn't hurt.
            "NSRequiresAquaSystemAppearance": False,
            **sparklePlist,
        }
        return {
            "data_files": [str(f) for f in self.dataFiles],
            "options": {
                "py2app": {
                    "plist": infoPList,
                    "iconfile": str(self.icnsFile),
                    "app": [str(self.mainPythonScript)],
                    "frameworks": [
                        *sparkleFrameworks,
                        *self.otherFrameworks,
                    ],
                    "excludes": [
                        # Excluding setuptools is a workaround for a problem in
                        # py2app -
                        # https://github.com/ronaldoussoren/py2app/issues/531 -
                        # and despite a couple of spuriously declared
                        # transitive dependencies on it
                        # (https://github.com/zopefoundation/zope.interface/issues/339,
                        # https://github.com/twisted/incremental/issues/141) we
                        # don't actually need it
                        "setuptools",
                    ],
                    # Workaround for
                    # https://github.com/ronaldoussoren/py2app/issues/546 -
                    # this needs validation to see if explicitly *including*
                    # Tcl.framework and Tk.framework does result in them
                    # getting signed properly, rather than blowing up in
                    # py2app's codesign_adhoc.
                    "dylib_excludes": [str(each) for each in dylibExcludes],
                }
            },
        }



"""
A script to replace setuptools' `setup.py`, so we can invoke py2app directly.
"""

if __name__ == '__main__':
    import sys

    sys.path.append(".")
    sys.argv[0] = "encrust-setup.py"

    from encrust_setup import description  # type:ignore[import-not-found]
    from setuptools import setup  # type:ignore[import-untyped]

    setup(**description.setupOptions())



import sys
from functools import wraps
from getpass import getpass
from os import environ
from os.path import abspath
from typing import Any, Callable, Concatenate, Coroutine, Generator, ParamSpec, TypeVar

import click
from twisted.internet.defer import Deferred
from twisted.internet.task import react
from twisted.python.failure import Failure

from ._build import AppBuilder
from ._spawnutil import c
from .api import AppDescription

P = ParamSpec("P")
R = TypeVar("R")


async def configuredBuilder() -> AppBuilder:
    """
    Make an AppBuilder out of the local configuration.
    """
    lines = await c.python(
        "-m",
        "encrust._dosetup",
        "--name",
        "--version",
        workingDirectory=abspath("."),
    )
    name, version = lines.output.decode("utf-8").strip().split("\n")
    name = environ.get("ENCRUST_APP_NAME", name)
    version = environ.get("ENCRUST_APP_VERSION", version)
    return AppBuilder(name=name, version=version)


def reactorized(
    c: Callable[
        Concatenate[Any, P],
        Coroutine[Deferred[object], Any, object]
        | Generator[Deferred[object], Any, object],
    ],
) -> Callable[P, None]:
    """
    Wrap an async twisted function for click.
    """

    @wraps(c)
    def forclick(*a, **kw) -> None:
        def r(reactor: Any) -> Deferred[object]:
            async def ar():
                try:
                    await c(reactor, *a, **kw)
                except Exception:
                    print(Failure().getTraceback())

            return Deferred.fromCoroutine(ar())

        react(r, [])

    return forclick


@click.group()
def main() -> None:
    """
    Utility for building, signing, and notarizing macOS applications.
    """


@main.command()
@reactorized
async def signable(reactor: Any) -> None:
    """
    (Debugging): print a list of every signable path in an already-built bundle.
    """
    builder = await configuredBuilder()
    for p in builder.signablePaths():
        print(p.path)


@main.command()
@reactorized
async def sign(reactor: Any) -> None:
    """
    (Debugging): Just locally codesign (and do not notarize) an already-built app.
    """
    builder = await configuredBuilder()
    await builder.signApp()


@main.command()
@reactorized
async def fatten(reactor: Any) -> None:
    """
    Ensure that all locally installed shared objects are fat binaries (i.e.
    universal2 wheels).
    """
    builder = await configuredBuilder()
    await builder.fattenEnvironment()


@main.command()
@reactorized
async def build(reactor: Any) -> None:
    """
    Build the application.
    """
    builder = await configuredBuilder()
    await builder.build()
    await builder.signApp()


@main.command()
@reactorized
async def devalias(reactor: Any) -> None:
    """
    Build an app bundle that uses a symlink into the development copy of the
    source code, suitable only for local development, but a lot faster than
    rebuilding all the time.

    @see: U{py2app alias mode
        <https://py2app.readthedocs.io/en/latest/tutorial.html#development-with-alias-mode>}
    """
    builder = await configuredBuilder()
    await builder.build("--alias")


@main.command()
@reactorized
async def release(reactor: Any) -> None:
    """
    Build the application.
    """
    builder = await configuredBuilder()
    await builder.release()


def loadDescription() -> AppDescription:
    """
    Load the description of the project from C{encrust_setup.py}.
    """
    sys.path.append(".")
    import encrust_setup  # type:ignore[import-not-found]

    desc = encrust_setup.description
    return desc


@main.command()
@reactorized
async def getsparkle(reactor: Any) -> None:
    """
    Download the Sparkle framework used by the current project.
    """
    # TODO: should probably use something like Cocoapods to actually fetch
    # frameworks so that this generalizes a bit.  But I would have to learn how
    # Cocoapods work for that.
    description = loadDescription()
    if description.sparkleData is None:
        print("Sparkle not specified, not downloading.")
        sys.exit(1)

    await description.sparkleData.sparkleFramework.download()


@main.command()
@reactorized
async def appcastify(reactor: Any) -> None:
    """
    Update, sign, and deploy the Sparkle appcast for the current application.

    Note that this currently must be manually done *after* `encrust release`,
    but we should probably integrate it into that process.
    """
    description = loadDescription()
    if description.sparkleData is None:
        print("Sparkle not specified, not generating appcast.")
        sys.exit(1)
    await description.sparkleData.deploy()


@main.command()
@reactorized
async def auth(reactor: Any) -> None:
    """
    Authenticate to the notarization service with an app-specific password from
    https://appleid.apple.com/account/manage
    """
    builder = await configuredBuilder()
    sign = await builder.signingConfiguration()
    newpw = getpass(f"Paste App-Specific Password for {sign.appleID} and hit enter: ")
    await builder.authenticateForSigning(newpw)
    print("Authenticated!")


@main.command()
@reactorized
async def configure(reactor: Any) -> None:
    """
    Configure this tool.
    """
    print(
        """
    TODO: this tool should walk you through configuration.

    For now:
    0. First, set up a Python project built using `py2app`.
        a. make a virtualenv
        b. `pip install` your dependencies
        c. `pip install encrust`

    1. enroll in the Apple Developer program at https://developer.apple.com/account
    2. download Xcode.app from https://apps.apple.com/us/app/xcode/id497799835?mt=12
    3. launch Xcode,
        a. open Preferences -> Accounts
        b. hit '+' to log in to the Apple ID you enrolled in
           the developer program with
        c. click "manage certificates"
        d. click "+"
        e. click "Developer ID Application"
    4. run `security find-identity -v -p codesigning`
    5. look for a "Developer ID Application" certificate in the list
    6. edit ~/.encrust.json to contain an object like this:

        {
            "identity": /* the big hex ID from find-identity output */,
            "teamID": /* the thing in parentheses in find-identity output */,
            "appleID": /* the email address associated with your apple developer account */,
            "profile": /* an arbitrary string you've selected */
        }
    7. go to https://appleid.apple.com/account/manage and log in
    8. click "App-Specific Passwords"
    9. run `encrust auth` and paste the app password before closing the window
    10. run `encrust release`
    11. upload dist/<YourApp>-<YourVersion>.release.app.zip somewhere on the web.
    """
    )



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Iterable

from twisted.internet.defer import Deferred
from twisted.python.filepath import FilePath

from ._spawnutil import c, parallel


@dataclass
class CodeSigner:
    bundle: FilePath[str]
    codesigningIdentity: str
    entitlements: FilePath[str]
    progress: dict[FilePath[str], Deferred[None]] = field(default_factory=dict)

    async def sign(self) -> None:
        active = 0
        async def signOneFile(fileToSign: FilePath[str]) -> None:
            """
            Code sign a single file.
            """
            nonlocal active
            fileStr = fileToSign.path
            entitlementsStr = self.entitlements.path
            print(f"code signing (|| {active}/{len(self.progress)}) {fileToSign}", flush=True)
            allChildren = []
            for eachMaybeChild in self.progress:
                if fileToSign in eachMaybeChild.parents():
                    allChildren.append((eachMaybeChild, self.progress[eachMaybeChild]))
            self.progress[fileToSign] = Deferred()
            for pn, toAwait in allChildren:
                print(f"waiting for {pn.path!r}…")
                await toAwait
                print(f"done waiting for {pn.path!r}!")
            try:
                active += 1
                await c.codesign(
                    "--sign",
                    self.codesigningIdentity,
                    "--entitlements",
                    entitlementsStr,
                    "--force",
                    "--options",
                    "runtime",
                    fileStr,
                )
            finally:
                self.progress.pop(fileToSign).callback(None)
                active -= 1
            print(f"finished signing (|| {active}/{len(self.progress)}) {fileToSign}", flush=True)

        async for signResult in parallel(
            (signOneFile(p) for p in signablePathsIn(self.bundle))
        ):
            pass


MACH_O_MAGIC = {
    b"\xca\xfe\xba\xbe",
    b"\xcf\xfa\xed\xfe",
    b"\xce\xfa\xed\xfe",
    b"\xbe\xba\xfe\xca",
    b"\xfe\xed\xfa\xcf",
    b"\xfe\xed\xfa\xce",
}


def hasMachOMagic(p: FilePath[str]) -> bool:
    with p.open("r") as f:
        magic = f.read(4)
        return magic in MACH_O_MAGIC


def signablePathsIn(topPath: FilePath[str]) -> Iterable[FilePath[str]]:
    """
    What files need to be individually code-signed within a given bundle?
    """
    built = []
    for p in topPath.walk(lambda subp: (not subp.islink() and subp.isdir())):
        if p.islink():
            continue
        ext = p.splitext()[-1]
        if p.isfile():
            if ext == "":
                if hasMachOMagic(p):
                    built.append(p)
            if ext in {".so", ".dylib", ".a"}:
                built.append(p)
        if p.isdir():
            if ext in {".framework", ".app", ".xpc"}:
                built.append(p)
    return reversed(built)


async def notarize(
    *,
    archivePath: FilePath[str],
    applicationPath: FilePath[str],
    appleID: str,
    teamID: str,
    notarizeProfile: str,
) -> None:
    """
    Submit the signed bundle for notarization, wait for success, then notarize
    it.
    """
    await c.xcrun(
        "notarytool",
        "submit",
        archivePath.path,
        f"--apple-id={appleID}",
        f"--team-id={teamID}",
        f"--keychain-profile={notarizeProfile}",
        "--wait",
    )
    await c.xcrun("stapler", "staple", applicationPath.path)



"""
Future work:

- integrate cocoapods
"""

from __future__ import annotations

import sys
from dataclasses import dataclass
from json import load
from os.path import abspath, expanduser
from typing import Iterable

from twisted.python.filepath import FilePath
from twisted.python.modules import getModule

from ._architecture import fixArchitectures, validateArchitectures
from ._signing import CodeSigner, notarize, signablePathsIn
from ._spawnutil import c
from ._zip import createZipFile


@dataclass
class AppSigner:
    notarizeProfile: str
    appleID: str
    teamID: str
    identityHash: str
    entitlementsPath: FilePath[str] = getModule(__name__).filePath.sibling(
        "required-python-entitlements.plist"
    )


@dataclass
class AppBuilder:
    """
    A builder for a particular application
    """

    name: str
    version: str
    _signer: AppSigner | None = None

    async def signingConfiguration(self) -> AppSigner:
        """
        Load the global signing configuration.
        """
        if self._signer is None:
            with open(expanduser("~/.encrust.json")) as f:
                obj = load(f)
            self._signer = AppSigner(
                identityHash=obj["identity"],
                notarizeProfile=obj["profile"],
                appleID=obj["appleID"],
                teamID=obj["teamID"],
            )
        return self._signer

    async def release(self) -> None:
        """
        Execute the release end to end; build, sign, archive, notarize, staple.
        """
        await self.fattenEnvironment()
        await self.build()
        archOK = await validateArchitectures([self.originalAppPath()], True)
        if not archOK:
            raise RuntimeError()
        await self.signApp()
        await self.notarizeApp()

    async def fattenEnvironment(self) -> None:
        """
        Ensure the current virtualenv has all universal2 "fat" binaries.
        """
        pathEntries = [FilePath(each) for each in sys.path if each]
        needsFattening = not await validateArchitectures(pathEntries)
        if not needsFattening:
            print("already ok")
            return
        await fixArchitectures()
        stillNeedsFattening = not await validateArchitectures(pathEntries, True)
        if stillNeedsFattening:
            raise RuntimeError(
                "single-architecture binaries still exist after fattening: {stillNeedsFattening}"
            )
        print("all relevant binaries now universal2")

    def archivePath(self, variant: str) -> FilePath[str]:
        """
        The path where we should archive our zip file.
        """
        return FilePath("dist").child(f"{self.name}-{self.version}.{variant}.app.zip")

    async def archiveApp(self, variant: str) -> FilePath[str]:
        """ """
        archivedAt = self.archivePath(variant)
        await createZipFile(archivedAt, self.originalAppPath())
        return archivedAt

    async def build(self, *options: str) -> None:
        """
        Invoke py2app to build a copy of the application, with the given py2app
        options.
        """
        await c.python(
            "-m",
            "encrust._dosetup",
            "py2app",
            *options,
            workingDirectory=abspath("."),
        )

    async def authenticateForSigning(self, password: str) -> None:
        """
        Prompt the user to authenticate for code-signing and notarization.
        """
        sign = await self.signingConfiguration()
        await c.xcrun(
            "notarytool",
            "store-credentials",
            sign.notarizeProfile,
            "--apple-id",
            sign.appleID,
            "--team-id",
            sign.teamID,
            "--password",
            password,
        )

    def originalAppPath(self) -> FilePath[str]:
        """
        A L{FilePath} pointing at the application (prior to notarization).
        """
        return FilePath("./dist").child(self.name + ".app")

    def signablePaths(self) -> Iterable[FilePath]:
        return signablePathsIn(self.originalAppPath())

    async def signApp(self) -> None:
        """
        Find all binary files which need to be signed within the bundle and run
        C{codesign} to sign them.
        """
        sign = await self.signingConfiguration()
        signer = CodeSigner(
            self.originalAppPath(),
            sign.identityHash,
            sign.entitlementsPath,
        )
        await signer.sign()

    async def notarizeApp(self) -> None:
        """
        Submit the built application to Apple for notarization and wait until we
        have seen a response.
        """
        sign = await self.signingConfiguration()
        preReleasePath = await self.archiveApp("for-notarizing")
        await notarize(
            appleID=sign.appleID,
            teamID=sign.teamID,
            archivePath=preReleasePath,
            applicationPath=self.originalAppPath(),
            notarizeProfile=sign.notarizeProfile,
        )
        await self.archiveApp("release")
        preReleasePath.remove()
